#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "terrain.h"

// scale is (# pixels) / width
terrain_t *trn_new(int w, int h, float scale)
{
	terrain_t *map;
	
	if (!w || !h) return NULL;

	map = (terrain_t*)malloc(sizeof(terrain_t));

	if (!map) return NULL;

	map->w = w;
	map->h = h;
	map->xs = map->ys = scale;

	map->terrain = (float*)malloc(w * h * sizeof(float));

	if (!(map->terrain))
	{
		free(map);
		return NULL;
	}

	memset(map->terrain, 0, w * h * sizeof(float));

	return map;
}

void trn_free(terrain_t *t)
{
	if (t)
	{
		if (t->terrain) free(t->terrain);
		free(t);
	}
}

// raise/lower the whole landscape
void trn_raise(terrain_t *t, float h)
{
	float *f;

	if (!t) return;

	for (f = t->terrain; f < t->terrain + t->w * t->h; f++)
		*f += h;
}

// add a cone to a landscape
// cone is centered on x, y; has height h; has base radius r
void trn_addcone(terrain_t *t, float x, float y, float r, float h)
{
	int i, j;
	float s, k = r * r;

	if (!t) return;

#define X	((i - (t->w/2))/t->xs)
#define Y	((j - (t->h/2))/t->ys)

	for (i = 0; i < t->w; i++)
	for (j = 0; j < t->h; j++)
	{
		if ((s = (X - x) * (X - x) + (Y - y) * (Y - y)) < k)
		{
			s = sqrt(s);
			t->terrain[i + j * t->w] += (1 - (s/r)) * h;
		}
	}

#undef X
#undef Y
}

void trn_addgaussian(terrain_t *t, float x, float y, float dev, float h)
{
	int i, j;
	float s;

	if (!t) return;

#define X	((i - (t->w/2))/t->xs)
#define Y	((j - (t->h/2))/t->ys)

	for (i = 0; i < t->w; i++)
	for (j = 0; j < t->h; j++)
	{
		s = (X - x) * (X - x) + (Y - y) * (Y - y);

		t->terrain[i + j * t->w] += exp(-s / (dev*dev)) * h;
	}

#undef X
#undef Y
}

// tilt the landscape at gradient h centered on a in direction r
void trn_tilt(terrain_t *t, float ax, float ay, float rx, float ry, float h)
{
	int i, j;
	float r;

	if (!t) return;

	r = sqrt(rx * rx + ry * ry);
	rx /= r; ry /= r;

#define X	((i - (t->w/2))/t->xs)
#define Y	((j - (t->h/2))/t->ys)

	for (i = 0; i < t->w; i++)
	for (j = 0; j < t->h; j++)
	{
		t->terrain[i + j * t->w] += (float)((X - ax) * rx + (Y - ay) * ry) * h;
	}

#undef X
#undef Y
}

// fault landscape up by h on side of landscape in r direction from a, down
// on other side
void trn_fault(terrain_t *t, float ax, float ay, float rx, float ry, float h)
{
	int i, j;
	float r;

	if (!t) return;

	r = sqrt(rx * rx + ry * ry);
	rx /= r; ry /= r;

#define X	((i - (t->w/2))/t->xs)
#define Y	((j - (t->h/2))/t->ys)

	for (i = 0; i < t->w; i++)
	for (j = 0; j < t->h; j++)
	{
		t->terrain[i + j * t->w] += ((X - ax) * rx + (Y - ay) * ry) > 0 ? h : -h;
	}

#undef X
#undef Y
}

// normalises terrain to lie between 0 and h
void trn_normalise(terrain_t *t, int h)
{
	float min, max, *f;

	if (!t) return;

	// determine min/max values
	min = max = *(t->terrain);
	
	for (f = t->terrain; f < t->terrain + t->w * t->h; f++)
	{
		if (*f < min) min = *f;
		else if (*f > max) max = *f;
	}

	for (f = t->terrain; f < t->terrain + t->w * t->h; f++)
		*f = h * (*f - min) / (max - min);
}

// remove all terrain with h < 0
void trn_cutoff(terrain_t *t)
{
	float *a;
	
	if (!t) return;

	for (a = t->terrain; a < t->terrain + t->w * t->w; a++)
		if (*a < 0) *a = 0;
}

// draw a line into the terrain
void trn_drawline(terrain_t *t, float x0, float y0, float h0, float x1, float y1, float h1)
{
	float i, j, k, di, dj, dk, l;
	int n;

	if (!t) return;

	// convert i, j, k etc into pixel dimensions
	di = (x1 - x0) * t->xs;
	dj = (y1 - y0) * t->ys;
	dk = h1 - h0;

	l = sqrt(di * di + dj * dj);
	di /= l;
	dj /= l;
	dk /= l;

	for (n = 0, i = x0 * t->xs, j = y0 * t->ys, k = h0; n < (int)l; i += di, j += dj, k += dk, n++)
	{
		t->terrain[(int)(i + t->w / 2) +(int)(j + t->h / 2) * t->w] = k;
	}
}

// figure out the landscape's min/max altitudes
void trn_minmax(terrain_t *t, float *min, float *max)
{
	float *a;
	
	if (!t) return;
	
	for (a = t->terrain, *max = *min = *a; a < t->terrain + t->w * t->h; a++)
	{
		if (*a < *min) *min = *a;
		else if (*a > *max) *max = *a;
	}
}

// output the terrain as a scaled byte array
unsigned char *trn_bytearray(terrain_t *t, int *w, int *h, float *min, float *max)
{
	float *f;
	unsigned char *data, *u;
	
	if (!t) return NULL;

	*w = t->w;
	*h = t->h;

	data = (unsigned char*)malloc(t->w * t->h);

	if (!data) return NULL;

	// determine min/max values
	*min = *max = *(t->terrain);
	
	for (f = t->terrain; f < t->terrain + t->w * t->h; f++)
	{
		if (*f < *min) *min = *f;
		else if (*f > *max) *max = *f;
	}

	// write out bytes
	for (f = t->terrain, u = data; f < t->terrain + t->w * t->h; f++, u++)
		*u = (unsigned char)(((*f - *min) / (*max - *min)) * 255.0);

	return data;
}

#define MAXVAL	65535

void trn_writepgm(terrain_t *t, FILE *fp)
{
	float *f;
	float min, max;

	if (!t || !f) return;
	
	fprintf(fp, "P2\n%d\n%d\n%d\n", t->w, t->h, MAXVAL);

	// determine min/max values
	min = max = *(t->terrain);
	
	for (f = t->terrain; f < t->terrain + t->w * t->h; f++)
	{
		if (*f < min) min = *f;
		else if (*f > max) max = *f;
	}
	
	for (f = t->terrain; f < t->terrain + t->w * t->h; f++)
	{
		fprintf(fp, "%d\n", (int)((float)MAXVAL * (*f - min)/(max - min)));
	}
}

void trn_writecarray(terrain_t *t, FILE *fp, char *name)
{
	int i, j;

	if (!t) return;

	fprintf(fp, "float %s[%d][%d] = {\n", name ? "alt" : name, t->h, t->w);

	for (i = 0; i < t->h; i++)
	{
		fprintf(fp, "    {");
		for (j = 0; j < t->w; j++)
		{
			fprintf(fp, "%f%s", *(t->terrain + j * t->w + i),
					j == t->w - 1 ? "" : ", ");
		}
		fprintf(fp, "}\n");
	}
	
	fprintf(fp, "  }\n");
}
